#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "figbuild.h"
#include "aux.h"
#include "ggw.h"

static char rcsid[]="$Id: figbuild.c,v 2.17 2002/03/09 18:32:55 apian Exp apian $";
static char version[]="$Revision: 2.17 $";

/* 'scanargs' command line processing was adapted from URT's rle library */

extern char *strsave();
extern	int	debug;

	float	scale=1.0;
	float	origin[2]={0,0};
	int	inside_flag=0;
	int	no_orient=0;

static	int	debug=0;
static	int	alltext=0;

GGW_io *goi=NULL;

/*--------------------------------------------------------------------------------------------------------------*/
/*
*/
int count_rargs( char *str )		/* count radiance arguments in string */
{
	register int i=1, wsf=0;
#ifdef DEBUG
	fprintf(stderr,"count_rargs: \"%s\" ...",str );
#endif

	while( *str ) {
	  if( isspace(*str) ) {		/* is white space */
	    wsf=1;
	  } else {			/* no white space */
	    if( wsf ) i++;		/* following a white space, no another parameter */
	    wsf=0;
	  }
	  str++;
	}
#ifdef DEBUG
	fprintf(stderr,"%d\n", i );
#endif
	
	return(i);
}

/*--------------------------------------------------------------------------------------------------------------*/
int read_fig( char *name )			/* liest die Koordinaten des xfig-Files */
						/* fuehrt enter_wall_segment() aus       */
{
	FILE *fp;
	char string[1024];
	int v[2];	/* fig version */
	int npoints, q1, q2, obj,subtype;
	float *x, *y;
	register int i, n;
	float upi, fummy;
	int linestyle, thickness, pen_color, area_fill, dummy, compound=0, metric=1, font_flags;
	int fill_color, joinstyle, capstyle, radius;
	h_segment *hs;

	hs = h_seg_new();

 	if ((fp=fopen(name,"r"))==NULL) {
        	fprintf(stderr,"read_fig: cannot open %s\n",name);
        	return(-1);
      	}

	fscanf(fp,"%[^\n]",string); 		/* read one line */
	fscanf(fp,"%*[\n\r]");

	if( !strncmp(string,"#FIG",4) ) {			/* check FIG version */
	  sscanf(string,"%*s %d.%d",v,v+1);
	  if( v[0]<3 ) {
	        fprintf(stderr,"read_fig: cannot read FIG versions before 3.1 in file %s\n",name);
		goto fail;
	  }
	  while(1){						 /* read config lines */
	    if( EOF == fscanf(fp,"%[^\n]%*[\n\r]",string) ) {
	      fprintf(stderr,"read_fig: can't find upi line in %s\n",name);
	      goto fail;
	    }
	    if( !strcmp(string,"Metric") ) metric=1;		/* decide metric/inches */
	    if( !strcmp(string,"Inches") ) metric=0;
	    if( 2==sscanf(string,"%g %d",&upi,&q2) ) break;	/* look for first line with 2 ints */
	  }
 	  if( upi!= 1200 ) {
	    fprintf(stderr,"read_fig: upi!=1200, maybe input file %s is gorked\n",name);
	  }
	} else {
	  fprintf(stderr,"read_fig: can't gork first line in %s\n",name);
	  goto fail;
	}

	if( metric ) upi=450;					/* go metric */

	if(debug) fprintf(stderr,"read_fig: metric=%d upi=%g\n",metric,upi);

	if( 1.0!=scale ) {
	  fprintf(stderr,"read_fig: using scale %g\n",scale);
	  upi *= scale;
	}

	while( EOF!=fscanf(fp,"%d",&obj) ) {			/* try to read object identifier */

	  switch(obj){

	    case(0):						/* USER DEFINED COLOR */
		fscanf(fp,"%d",&pen_color);			/* color number */
		fscanf(fp,"%*s");				/* hex string */
		fprintf(stderr,"read_fig: ignored user defined color %d\n",pen_color);
		break;

	    case(2):						/* POLYLINE */
		fscanf(fp,"%d",&subtype);
		switch(subtype) {
		  case(1):					/* polyline */
		  case(2):					/* box */
		  case(3):					/* polygon */
		  case(4):					/* arc-box */
		  case(5):					/* image-box */
			fscanf(fp,"%d",&linestyle);		/* linestyle */
			fscanf(fp,"%d",&thickness);		/* thickness */
			fscanf(fp,"%d",&pen_color);		/* pencolor */
			fscanf(fp,"%d",&fill_color);		/* fill_color */
			fscanf(fp,"%d",&dummy);			/* depth */
			fscanf(fp,"%d",&dummy);			/* pen_style */
			fscanf(fp,"%d",&area_fill);		/* area_fill */
			fscanf(fp,"%*g");			/* style_val */
			fscanf(fp,"%d",&joinstyle);		/* join style */
			fscanf(fp,"%d",&capstyle);		/* cap style */
			fscanf(fp,"%d",&radius);		/* radius for arc-boxes */
			fscanf(fp,"%d %d",&q1,&q2);		/* arrow parameter */
			if( q1 || q2 ) {
			  fprintf(stderr,"read_fig: don't set arrows on any polygon line.\n");
			  goto fail;
			}
			fscanf(fp,"%d",&npoints);		/* number of points */

			/* maybe  npoints--  for images ?? */

			if( subtype==5 ) {			/* image box */
			  if( npoints!=5 ) {
			    fprintf(stderr,"read_fig: strange number of ppoints for picture.\n");
			    goto fail;
			  }
			  fscanf(fp,"%*d %s",string);		/* orientation and filename */
 			  fprintf(stderr,"read_fig: we ignore imported [%s] picture \"%s\"\n", fig_colorname(pen_color), string);
			}


			if( debug )
			  fprintf(stderr,"read_fig: POLYLINE color=\"%s\" style=\"%s\": %d coors\n",
				fig_colorname(pen_color), fig_linestylename(linestyle), npoints);

			x= (float*) tmalloc(sizeof(float)*npoints);
			y= (float*) tmalloc(sizeof(float)*npoints);
			for(i=0;i<npoints;++i) {		/* read coordinates */
			  if( 2 != fscanf(fp,"%g %g ",x+i,y+i) ) {
			    fprintf(stderr,"read_fig: EOF in coordinate read from \"%s\"\n", name);
			    goto fail;
			  }
			  x[i] = x[i]/upi - origin[0];
			  y[i] = -(y[i]/upi - origin[1]);
			}

	 		sprintf(string,"figseg-%s-%s",
				fig_colorname(pen_color), fig_linestylename(linestyle) );

			if( !(hs->vs= v_segs_lookup( string ) ) ) {
			  fprintf(stderr,"read_fig: no v_seg \"%s\"\n", string );
			  exit(2);
			}

			if( subtype==4 ) 			/* arc box */
			  fprintf(stderr,"... ignoring radius for arc-box\n");


			hs->type = HSEG_TYPE_POLYGON;	/* fill in structure */
			hs->linestyle = linestyle;
			hs->thickness = thickness;
			hs->filled = (area_fill==(-1) ? 0x10:0x01 );
			hs->joinstyle = joinstyle;
			hs->npoints = npoints;
			hs->x = x;
			hs->y = y;
			segment_io( hs );
			free(hs->x);
			free(hs->y);

		  break;


#ifdef IMAGE_XTRA
	/*
		we try to gork imported images, since they appear frequently
	*/
		  case(5):					/* imported picture bounding box */
			fscanf(fp,"%*d");		/* linestyle */
			fscanf(fp,"%*d");		/* thickness */
			fscanf(fp,"%*d");		/* pencolor */
			fscanf(fp,"%*d");		/* fill_color */
			fscanf(fp,"%*d");		/* depth */
			fscanf(fp,"%*d");		/* pen_style */
			fscanf(fp,"%*d");		/* area_fill */
			for(i=0;i<4;++i) fscanf(fp,"%*s");	/* read other polygon parameters */
			fscanf(fp,"%d %d",&q1,&q2);		/* arrow parameter */
			if( q1 || q2 ) {
			  fprintf(stderr,"read_fig: don't set arrows on any picture.\n");
			  goto fail;
			}
			fscanf(fp,"%d",&npoints);		/* number of points */
			if( npoints!=5 ) {
			  fprintf(stderr,"read_fig: strange number of ppoints for picture.\n");
			  goto fail;
			}
			fscanf(fp,"%*d %s",string);		/* orientation and filename */
 			fprintf(stderr,"read_fig: we ignore imported picture \"%s\".\n",string);
			for(i=0;i<npoints;++i)  		/* coordinates */
			  fscanf(fp," %*d %*d "); 
		  break;
#endif
	/*
		all other primitives we don't know how to handle .
	*/

		  default:
			fprintf(stderr,"read_fig: can't gork polyline subtype %d \n",subtype);
			goto fail;
		}
		break;

	    case(5):						/* ARC */
		fscanf(fp,"%d",&subtype);
		switch(subtype) {
		  case(0):					/* pie wegded */	
		  case(1):					/* open arc */
			fscanf(fp,"%d",&linestyle);		/* linestyle */
			fscanf(fp,"%d",&thickness);		/* thickness */
			fscanf(fp,"%d",&pen_color);		/* pencolor */
			fscanf(fp,"%d",&fill_color);		/* fill_color */
			fscanf(fp,"%d",&dummy);			/* depth */
			fscanf(fp,"%d",&dummy);			/* pen_style */
			fscanf(fp,"%d",&area_fill);		/* area_fill */
			fscanf(fp,"%*g");			/* style_val */
			fscanf(fp,"%*d");			/* cap_style */
			fscanf(fp,"%*d");			/* direction */
			fscanf(fp,"%d %d",&q1,&q2);		/* arrow parameter */
			if( q1 || q2 ) {
			  fprintf(stderr,"read_fig: don't set arrows on any ARC.\n");
			  goto fail;
			}
			fscanf(fp,"%*g %*g");			/* center of the arc */
			fscanf(fp,"%*d %*d");			/* x,y */
			fscanf(fp,"%*d %*d");			/* x,y */
			fscanf(fp,"%*d %*d");			/* x,y */
			break;
		   default:
			fprintf(stderr,"read_fig: unkown ARC subtype %d\n",subtype);
			goto fail;
		}
		break;

	    case(4):						/* TEXT */
		x=(float*) tmalloc(sizeof(float));
		y=(float*) tmalloc(sizeof(float));
		fscanf(fp,"%d",&subtype);	/* justified */
		fscanf(fp,"%d",&dummy);		/* color */
		fscanf(fp,"%d",&dummy);		/* depth */
		fscanf(fp,"%d",&dummy);		/* pen_style */
		fscanf(fp,"%d",&dummy);		/* font */
		fscanf(fp,"%*g");		/* font_size */
		fscanf(fp,"%*g");		/* angle */
		fscanf(fp,"%d",&font_flags);	/* font_flags */
		fscanf(fp,"%*g");		/* height */
		fscanf(fp,"%*g");		/* width */
		fscanf(fp,"%g %g",x, y);	/* position */
		fscanf(fp,"%*c");		/* blank */
		for( i=0; i<1024 ; ++i ) {
			fread( string+i, 1, 1, fp );
			if( i>=3 && !memcmp(string+i-3,"\\001",4 ) ) break;
		}
		string[i-3]= '\0';
		n= count_rargs( string );
#ifdef DEBUG_TEXT
		fprintf(stderr,"text position: %g %g \n",(float)x[0]/upi,(float)y[0]/upi );
		fprintf(stderr,"text: >%s< font_flags: %d\n", string, font_flags & 0x0f );
#endif
		if( alltext || (font_flags & 0x02) ) {	/* 'special' text is used for instances */
	          if(debug) fprintf(stderr,"instance %s\n", string );
		  fprintf(stdout,"void instance fi\n%d %s -t %g %g 0\n0\n0\n", 4+n, 
				string, x[0]/upi-origin[0],-(y[0]/upi-origin[1]) );
		}
		free(x);
		free(y);
		break;

	    case(3):						/* SPLINE */
		 fprintf(stderr,"read_fig: cannot gork SPLINE\n");
		 goto fail; 

	    case(1):						/* ELLIPSE */
		fscanf(fp,"%d",&subtype);
		switch(subtype) {
		  case(1):					/* ellipse / radius */
		  case(2):					/* ellipse / diameter */
		    fprintf(stderr,"read_fig: cannot gork ellipse subtype\n");
		    goto fail; 
		  case(3):					/* circle / radius */
		  case(4):					/* circle / diameter */
			{
	 		int cx, cy, rx, ry;
			fscanf(fp,"%d",&linestyle);		/* linestyle */
			fscanf(fp,"%d",&thickness);		/* thickness */
			fscanf(fp,"%d",&pen_color);		/* pencolor */
			fscanf(fp,"%d",&fill_color);		/* fill_color */
			fscanf(fp,"%d",&dummy);			/* depth */
			fscanf(fp,"%d",&dummy);			/* pen_style */
			fscanf(fp,"%d",&area_fill);		/* area_fill */
			fscanf(fp,"%g",&fummy);			/* style_val */
			fscanf(fp,"%d",&dummy);			/* direction */
			fscanf(fp,"%g",&fummy);			/* angle */
			fscanf(fp,"%d %d",&cx,&cy);
			fscanf(fp,"%d %d",&rx,&ry);
			fscanf(fp,"%d %d",&dummy,&dummy);	/* start */
			fscanf(fp,"%d %d",&dummy,&dummy);	/* end */
#ifdef DEBUG
			fprintf(stderr,"cx=%d cy=%d rx=%d ry=%d %d\n",cx,cy,rx,ry,dummy);
#endif
	 		sprintf(string,"figseg-%s-%s",
				fig_colorname(pen_color), fig_linestylename(linestyle) );

			if( !(hs->vs= v_segs_lookup( string ) ) ) {
			  fprintf(stderr,"read_fig: no v_seg \"%s\"\n", string );
			  exit(2);
			}

			hs->type = HSEG_TYPE_CIRCLE;	/* fill in structure */
			hs->linestyle = linestyle;
			hs->thickness = thickness;
			hs->filled = (area_fill==(-1) ? 0x10:0x01 );
			hs->npoints = 2;
			hs->x = (float*) tmalloc(sizeof(float)*2);
			hs->y = (float*) tmalloc(sizeof(float)*2);
			hs->x[0]=  cx/upi-origin[0];
			hs->y[0]= -(cy/upi-origin[1]);
			hs->x[1]=  rx/upi;
			hs->y[1]=  ry/upi;

			segment_io( hs ); 
			free(hs->x);
			free(hs->y);
			}
		    break;
		  default:
			fprintf(stderr,"read_fig: can't gork ellipse subtype %d \n",subtype);
			goto fail;
		}
		break;
	    case(6):						/* COMPOUND */
		compound++;
		for(i=0;i<4;++i) fscanf(fp,"%*d ");		/* suck parameters */
	        break;
	    case(-6):
		compound--;
		if( compound<0 ) {
		  fprintf(stderr,"read_fig: compound counter went pessimistic\n");
		}
	        break;

	    default:
		fprintf(stderr,"read_fig: cannot gork object type %d \n",obj);
		goto fail;

	  }
	}

	return(0);
fail:
 	fclose(fp);
	return(-1);
}

/*------------------------------------------------------------------------------------------*/
main( int argc, char **argv)
{
	extern  FILE    *yyin;
	extern  char    yyfilename[];
	char *figfile=NULL;
	char *heightfile=NULL;
	char *format=NULL;
	int q;
	int version_flag=0;

	if( !scanargs(argc, argv,
		"% V%- d%- t%- i%- n%- f%-output_format!s h%-heightfile!s g%-figfile!s",
		&version_flag,
		&debug,
		&alltext,
		&inside_flag,
		&no_orient,
		&q, &format,
		&q, &heightfile,
		&q, &figfile
		) ) {
	    fprintf(stderr,"generate 3d RADIANCE geometry from 2d FIG floorplan\n");
	    fprintf(stderr,"\t-V\tdisplay version\n");
	    fprintf(stderr,"\t-d\tturn on debugging\n");
	    fprintf(stderr,"\t-t\taccept all text (else only text with 'special' attribute)\n");
	    fprintf(stderr,"\t-i\tall polygons face inward (default is outward)\n");
	    fprintf(stderr,"\t-n\tdon't check for polygon orientation\n");
	    fprintf(stderr,"\t-f output_format\tradiance (default) | rayshade\n");
	    fprintf(stderr,"see  http://www.pab-opto.de/radiance/toolbox/figbuild/figbuild.html  for more\n");
	    exit(4);
	}

	if( version_flag ) {
	  fprintf(stderr,"%s\n",version);
	  exit(0);
	}

	if( !figfile ) {
	  fprintf(stderr,"figfile missing (try: figbuild -h)\n");
	  exit(1);
	}

	if( format && !strcmp(format,"rayshade") ) 
	  goi= ggw_open( "-", "rayshade", NULL );
	else
	  goi= ggw_open( "-", "radiance", NULL );

	if( !goi) {
	  fprintf(stderr,"ggw open failed\n");
	  exit(1);
	}

	if( heightfile ) {
	  if( NULL == ( yyin = fopen( heightfile, "r" ) ) ) {
	    fprintf(stderr,"figbuild: can't open height input file %s\n", argv[1] );
	    exit(1);
	  }
	  strcpy( yyfilename, heightfile );
	  yyparse();
	}

	if( figfile ) {
	  if( read_fig( figfile ) ) exit(1);
	}

	ggw_close( goi );

	exit(0);

}
/*-----------------------------------------------------------------------------*/
