
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "figbuild.h"
#include "aux.h"
#include "ggw.h"

static char rcsid[]="$Id: io.c,v 1.14 2002/03/09 18:41:08 apian Exp apian $";

static	int cnt=0;

extern	int inside_flag;
extern	int no_orient;

extern	GGW_io *goi;
extern	int	debug;	

#define	NOTVOIDMAT(a)	( (a) && strlen(a) && strcmp(a,"void") )

/*----------------------------------------------------------------------------------------*/
void one_wall_poly( h_segment *hs, int i1, int i2 )
{
	ggwvec v1, v2, v3, v4;
	register v_segments *vs;
	register int j;

	if( !(vs= hs->vs) ) return;

	v1[0]=v4[0]=hs->x[i1];
	v1[1]=v4[1]=hs->y[i1];
	v2[0]=v3[0]=hs->x[i2];
	v2[1]=v3[1]=hs->y[i2];

	v3[2]=v4[2]=vs->gnd;
	for( j= vs->max -1 ; j>=0; --j ) {
	    v1[2]=v2[2]=vs->segs[j].height;
#ifdef DEBUG
	    fprintf(stderr,"one_wall_poly: \"%s\" from %g to %g\n",
			vs->segs[j].name, v1[2], v3[2] );
#endif
	    ggw_polygon( goi, vs->segs[j].name, 4, v1, v2, v3, v4 );
	    v3[2]=v4[2]= v1[2];
	}

	return;
}
/*----------------------------------------------------------------------------------------*/
static int segment_io_poly( h_segment *hs )
{
	register int i, mode;
	register int q, inside;
	register v_segments *vs;
	register extrusion_data *ex;
	ggwvec v1, v2;
	ggwvec *ggva;
	

/*
	check orientation
*/
	inside= no_orient ? 1 : h_seg_orientation( hs ) ^ inside_flag ;


/*
	do walls 
*/
	if( inside ) {
	  for( i=hs->npoints-1; i>0; --i ) one_wall_poly( hs, i, i-1 );
	} else {
	  for( i=0; i< hs->npoints-1; ++i ) one_wall_poly( hs, i, i+1 );
	}

/*
	do ceiling and floor 

	todo: add support for colorpict mapping (modify mat, calculate transform)
*/
	vs= hs->vs;

	if( (vs->filled & hs->filled) &&  (hs->npoints>2) ) {

#ifdef DEBUG
	  fprintf(stderr,"segment_io_poly: %s floor=\"%s\" ceil=\"%s\"\n",
		vs->name, vs->floormat, vs->ceilmat );
#endif
	  ggva= (ggwvec *) tmalloc( sizeof(ggwvec) * hs->npoints );
	  q= hs->npoints -1;

	  if( NOTVOIDMAT(vs->floormat) ) {	/* floor */
	    if( inside ) {
	      for( i=0; i< hs->npoints; ++i ) {
	        ggva[i][0] = hs->x[i];
	        ggva[i][1] = hs->y[i];
	        ggva[i][2] = vs->gnd;
	      }
	    }else{
	      for( i=0; i< hs->npoints; ++i ) {
	        ggva[i][0] = hs->x[q-i];
	        ggva[i][1] = hs->y[q-i];
	        ggva[i][2] = vs->gnd;
	      }
	    }
	    ggw_polygon_v( goi, vs->floormat, hs->npoints, ggva );
	  }
#ifdef DEBUG	
	  v_segs_dump( vs );
#endif
	  if( NOTVOIDMAT(vs->ceilmat) ) {	/* ceiling */
	    if( inside ) {
	      for( i=0; i< hs->npoints; ++i ) {
	        ggva[i][0] = hs->x[q-i];
	        ggva[i][1] = hs->y[q-i];
	        ggva[i][2] = vs->segs[0].height;
	      }
	    } else {
	      for( i=0; i< hs->npoints; ++i ) {
	        ggva[i][0] = hs->x[i];
	        ggva[i][1] = hs->y[i];
	        ggva[i][2] = vs->segs[0].height;
	      }
	    }
	    ggw_polygon_v( goi, vs->ceilmat, hs->npoints, ggva );
	  }

	  free(ggva);

	}
	
/*
	do extrusions
*/
	ex=hs->vs->ex;
	if( ex ) {
#ifdef DEBUG
	  fprintf(stderr,"segment_io_poly: mode %d\n",  hs->joinstyle );
#endif
	  switch( hs->joinstyle) {
		case(1): mode=GGW_CYL_ENDCYLS|GGW_CYL_ENDSPHS; break;
		case(2): mode=GGW_CYL_ENDCAPS; break;
		default: mode=GGW_NOP;
	  }
	  for( i=1; i<hs->npoints; ++i ) {
	    v1[0]=hs->x[i-1];
	    v1[1]=hs->y[i-1];
	    v1[2]=hs->vs->gnd;
	    v2[0]=hs->x[i];
	    v2[1]=hs->y[i];
	    v2[2]=hs->vs->gnd;
	    switch( hs->vs->ex->mode ) {
	      case( EXTRUSION_CYL ):
		mode &= GGW_CYL_ENDSPHS|GGW_CYL_ENDCAPS;
	        ggw_cylinder( goi, ex->mat, mode, v1, v2, hs->thickness * ex->scale_h );
		break;
	      case( EXTRUSION_BOX ):
		mode &= GGW_CYL_ENDCYLS|GGW_CYL_ENDCAPS;
	        ggw_box( goi, ex->mat, mode, v2, v1,
			hs->thickness * ex->scale_h, hs->thickness * ex->scale_v );
	        break;
	      default:
	    }
	  }
	}
}

/*----------------------------------------------------------------------------------------*/
static int segment_io_circle( h_segment *hs )
{
	register int j ;
	register float r;
	register v_segments *vs;
	ggwvec up={0,0,1};
	ggwvec down={0,0,-1};
	ggwvec v1, v2;

/*
	do walls 
*/
	vs= hs->vs;
	v1[0]=v2[0]= hs->x[0];		/* position and radius */
	v1[1]=v2[1]= hs->y[0];
	r=hs->x[1];
	for( j=0; j<vs->max; ++j ) {
	    if( j ) 
	    	v1[2]=vs->segs[j-1].height;
	    else
	    	v1[2]=vs->gnd;
	    v2[2]=vs->segs[j].height;
	    ggw_cylinder( goi, vs->segs[j].name, GGW_NOP, v1, v2, r );
	}

/*
	do ceiling and floor 
*/
#ifdef DEBUG
	fprintf(stderr,"segment_io_circle: %s floor=\"%s\" ceil=\"%s\"\n",
		vs->name, vs->floormat, vs->ceilmat );
#endif

	if( (vs->filled & hs->filled) && NOTVOIDMAT(vs->floormat) ){
	  v1[2]=vs->gnd;
	  ggw_disc( goi, vs->floormat, v1, inside_flag?up:down, r, 0.0 );
	}
	if( (vs->filled & hs->filled) && NOTVOIDMAT(vs->ceilmat)) {
	  v1[2]= vs->segs[0].height;
	  ggw_disc( goi, vs->ceilmat , v1, inside_flag?down:up, r, 0.0 );
	}

	
}
	
	
/*----------------------------------------------------------------------------------------*/
int segment_io( h_segment *hs )
{

	switch( hs->type ) {
	  case( HSEG_TYPE_POLYGON ):
	    segment_io_poly( hs );
	  break;
	  case( HSEG_TYPE_CIRCLE ):
	    segment_io_circle( hs );
	  break;
	  default:
		fprintf(stderr,"segment_io: unsupported type %d\n", hs->type );
	}
}

