%{
#include <stdio.h>
#include <string.h>

#include "figbuild.h"
#include "aux.h"

#define Float	double

/* $Header: /home/apian/r/gen/figbuild/RCS/yacc.y,v 1.5 2000/11/07 20:44:42 apian Exp apian $ */

char yyfilename[BUFSIZ];			/* Input filename */

extern FILE *yyin;				/* input file pointer */
extern int yylineno;				/* Current line # in file */

extern	float	scale;
extern char	*strsave();
extern	float	origin[];
extern	int	debug;

static	v_segments	*vs;		/* space for new segment */

static float	gnddef=0.0;			/* global defaults */
static char	ceildef[1024]="\0", floordef[1024]="\0";

%}
%union {
	char *c;
	int i;
	Float d;
}
%token <d> tFLOAT
%token <c> tSTRING 
%token <i> tTILT tDEFSEG tENDSEG tFLOORMAT tCEILMAT tFILL t2DSCALE
%token <i> tNOFILL tEXTRUDE tBOX tCYL
%token <i> tGROUND tORIGIN
%type <d> Float

%left '+' '-'
%left '*' '/' '%'
%left UMINUS
%right '^'
%%
Items		: /* empty */
		| Items Item
		;
Item		: DefineSegment | Globals
		;
Globals		: tFLOORMAT tSTRING
		{
		  strcpy( floordef, $2 );
		}
		| tCEILMAT tSTRING
		{
		  strcpy( ceildef, $2 );
		}
		| Float tGROUND
		{
		  gnddef=$1;
		}
		| t2DSCALE tFLOAT
		{
		  scale=$2;
		}
		| tORIGIN tFLOAT tFLOAT
		{
		  origin[0]=$2;
		  origin[1]=$3;
		}
DefineSegment	: SegStart SegDescs SegEnd
		;
SegStart	: tDEFSEG tSTRING
		{
		  vs = v_segs_new();
		  vs->gnd= gnddef;
		  vs->filled= 0x01;
		  strncpy( vs->floormat, floordef, STRING_MAX);
		  strncpy( vs->ceilmat, ceildef, STRING_MAX);
		  vs->name = strsave($2) ;
		  vs->max = 0;
	 	  vs->ex=NULL;
		}
		;
SegEnd		: tENDSEG
		{
#ifdef DEBUG
		fprintf(stderr,"segment \"%s\" completed with %d vsegs:\n",
				vs->name, vs->max );
		fprintf(stderr,"floor, ceil: \"%s\",\"%s\"\n", vs->floormat, vs->ceilmat );
#endif
		}
		;

SegDescs	: /* empty */
		| SegDescs SegDesc
		;
SegDesc		: Float tSTRING			/* arbitrary material, specified elsewhere */
		{
		  vs->segs[vs->max].height= $1;
		  vs->segs[vs->max].name  = strsave( $2 );
		  vs->segs[vs->max].tilt  = 0.0;
		  vs->max++;
		}
		| tFLOAT tTILT					/* tilt angle */
		{
		  if(vs->max>1) vs->segs[vs->max-1].tilt = $1;
		}
		| tFLOORMAT tSTRING
		{
		  strncpy( vs->floormat, $2, STRING_MAX);
		}
		| tCEILMAT tSTRING
		{
		  strncpy( vs->ceilmat, $2, STRING_MAX);
		}
		| Float tGROUND
		{
		  vs->gnd=$1;
		}
		| tNOFILL
		{
		  vs->filled= 0x00;
		}
		| tFILL
		{
		  vs->filled= 0x11;
		}
		| tEXTRUDE tBOX tSTRING tFLOAT tFLOAT
		{
		  vs->ex = (extrusion_data *) tmalloc(sizeof(extrusion_data));
		  vs->ex->mode = EXTRUSION_BOX;
		  strncpy( vs->ex->mat, $3, STRING_MAX);
		  vs->ex->scale_h = $4;
		  vs->ex->scale_v = $5;
		}
		| tEXTRUDE tCYL tSTRING tFLOAT
		{
		  vs->ex = (extrusion_data *) tmalloc(sizeof(extrusion_data));
		  vs->ex->mode = EXTRUSION_CYL;
		  strncpy( vs->ex->mat, $3, STRING_MAX);
		  vs->ex->scale_h = $4;
		}
		;
Float		: tFLOAT
		| '-' tFLOAT
		{ $$ = -$2; }
		| '+' tFLOAT
		{ $$ = $2; }
		;
%%
/*
 * Issue error message containing filename and line number, and exit.
 */
/*VARARGS1*/
yyerror(s, pat1, pat2)
char *s, *pat1, *pat2;
{
	fprintf(stderr,"Error: %s: line %d: ", yyfilename, yylineno);
	fprintf(stderr, s, pat1, pat2);
	if (*s && s[strlen(s) -1] != '\n')
		/* YACC doesn't put newlines on error messages. */
		fprintf(stderr,"\n");	
	fflush(stderr);
	exit(1);
}

